# -*- coding: utf-8 -*-
"""routing header for IPv6

``jspcap.protocols.internet.ipv6_route`` contains
``IPv6_Route`` only, which implements extractor for Routing
Header for IPv6 (IPv6-Route), whose structure is described
as below.

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Next Header  |  Hdr Ext Len  |  Routing Type | Segments Left |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                                                               |
.                                                               .
.                       type-specific data                      .
.                                                               .
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
# TODO: Implements extractor of all routing types.
from jspcap.corekit.infoclass import Info
from jspcap.protocols.protocol import Protocol


__all__ = ['IPv6_Route']


# IPv6 Routing Types
_ROUTING_TYPE = {
    0 : 'Source Route',                 # [RFC 5095] DEPRECATED
    1 : 'Nimrod',                       # DEPRECATED 2009-05-06
    2 : 'Type 2 Routing Header',        # [RFC 6275]
    3 : 'RPL Source Route Header',      # [RFC 6554]
  253 : 'RFC3692-style Experiment 1',   # [RFC 4727]
  254 : 'RFC3692-style Experiment 2',   # [RFC 4727]
  255 : 'Reserved',
}


class IPv6_Route(Protocol):
    """This class implements Routing Header for IPv6.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_ipv6_route -- read Routing Header for IPv6 (IPv6-Route)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Routing Header for IPv6'

    @property
    def alias(self):
        """Acronym of corresponding procotol."""
        return 'IPv6-Route'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipv6_route(self, length, extension):
        """Read Routing Header for IPv6.

        Structure of IPv6-Route header [RFC 8200][RFC 5095]:
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Next Header  |  Hdr Ext Len  |  Routing Type | Segments Left |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                                                               |
            .                                                               .
            .                       type-specific data                      .
            .                                                               .
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     route.next              Next Header
              1           8     route.length            Header Extensive Length
              2          16     route.type              Routing Type
              3          24     route.seg_left          Segments Left
              4          32     route.data              Type-Specific Data

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _hlen = self._read_unpack(1)
        _type = self._read_unpack(1)
        _left = self._read_unpack(1)
        _data = self._read_fileng(_hlen - 3)

        ipv6_route = dict(
            next = _next,
            length = _hlen + 1,
            type = _ROUTING_TYPE.get(_type, 'Unassigned'),
            sed_left = _left,
            data = _data,
        )

        if length is not None:
            length -= ipv6_route['length']
        ipv6_route['packet'] = self._read_packet(header=ipv6_route['length'], payload=length)

        if extension:
            self._protos = None
            return ipv6_route
        return self._decode_next_layer(ipv6_route, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipv6_route(length, extension))

    def __length_hint__(self):
        return 4



# -*- coding: utf-8 -*-
"""open shortest path first

`jspcap.protocols.link.ospf` contains `OSPF` only,
which implements extractor for Open Shortest Path First
(OSPF), whose structure is described as below.

 0                   1                   2                   3
 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|   Version #   |     Type      |         Packet length         |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                          Router ID                            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                           Area ID                             |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|           Checksum            |             AuType            |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Authentication                          |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                       Authentication                          |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.link.link import Link


__all__ = ['OSPF']


# OSPF Packet Types
TYPE = {
    1 : 'Hello',
    2 : 'Database Description',
    3 : 'Link State Request',
    4 : 'Link State Update',
    5 : 'Link State Acknowledgment',
}


# Authentication Types
AUTH = {
    0 : 'Null Authentication',
    1 : 'Simple Password',
    2 : 'Cryptographic Authentication',
}


class OSPF(Link):
    """This class implements Open Shortest Path First.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding protocol
        * layer -- str, `Link`
        * length -- int, header length of corresponding protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * type -- str, OSPF packet type

    Methods:
        * read_ospf -- read Open Shortest Path First

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor
        * _read_id_numbers -- read router and area IDs
        * _read_encrypt_auth -- read Authentication field when CA employed

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'Open Shortest Path First'

    @property
    def length(self):
        """Header length of current protocol."""
        return 24

    @property
    def type(self):
        """OSPF packet type."""
        return self._info.type

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ospf(self, length):
        """Read Open Shortest Path First.

        Structure of OSPF header [RFC 2328]:
            0                   1                   2                   3
            0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |   Version #   |     Type      |         Packet length         |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |                          Router ID                            |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |                           Area ID                             |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |           Checksum            |             AuType            |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |                       Authentication                          |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
           |                       Authentication                          |
           +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     ospf.version            Version #
              1           8     ospf.type               Type (0/1)
              2          16     ospf.len                Packet Length (header includes)
              4          32     ospf.router_id          Router ID
              8          64     ospf.area_id            Area ID
              12         96     ospf.chksum             Checksum
              14        112     ospf.autype             AuType
              16        128     ospf.auth               Authentication

        """
        if length is None:
            length = len(self)

        _vers = self._read_unpack(1)
        _type = self._read_unpack(1)
        _tlen = self._read_unpack(2)
        _rtid = self._read_id_numbers()
        _area = self._read_id_numbers()
        _csum = self._read_fileng(2)
        _autp = self._read_unpack(2)

        ospf = dict(
            version = _vers,
            type = TYPE.get(_type),
            len = _tlen,
            router_id = _rtid,
            area_id = _area,
            chksum = _csum,
            autype = AUTH.get(_autp) or 'Reserved',
        )

        if _autp == 2:
            ospf['auth'] = self._read_encrypt_auth()
        else:
            ospf['auth'] = self._read_fileng(8)

        length = ospf['len'] - 24
        ospf['packet'] = self._read_packet(header=24, payload=length)

        return self._decode_next_layer(ospf, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_ospf(length))

    def __length_hint__(self):
        return 24

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_id_numbers(self):
        """Read router and area IDs."""
        _byte = self._read_fileng(4)
        _addr = '.'.join([str(_) for _ in _byte])
        return _addr

    def _read_encrypt_auth(self):
        """Read Authentication field when Cryptographic Authentication is employed.

        Structure of Cryptographic Authentication [RFC 2328]:

             0                   1                   2                   3
             0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |              0                |    Key ID     | Auth Data Len |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |                 Cryptographic sequence number                 |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     -                       Reserved (must be zero)
              2          16     ospf.auth.key_id        Key ID
              3          24     ospf.auth.len           Auth Data Length
              4          32     ospf.auth.seq           Cryptographic Sequence Number

        """
        _resv = self._read_fileng(2)
        _keys = self._read_unpack(1)
        _alen = self._read_unpack(1)
        _seqn = self._read_unpack(4)

        auth = dict(
            key_id = _keys,
            len = _alen,
            seq = _seqn,
        )

        return auth



# -*- coding: utf-8 -*-
"""PCAP file headers

`jspcap.protocols.pcap` contains header descriptions for
PCAP files, including global header and frame header.

"""
from jspcap.protocols.pcap.frame import Frame
from jspcap.protocols.pcap.header import Header


__all__ = ['Frame', 'Header']



# -*- coding: utf-8 -*-
"""analyser for application layer

`jspcap.tools.analysis` works as a header quarter to analyse
and match application layer protocol. Then, call
corresponding modules and functions to extract the attributes.

"""
from jspcap.protocols.raw import Raw
from jspcap.utilities.decorators import seekset_ng
from jspcap.utilities.exceptions import ProtocolError

###############################################################################
# from jspcap.protocols.application.httpv1 import HTTPv1
# from jspcap.protocols.application.httpv2 import HTTPv2
###############################################################################


__all__ = ['Analysis']


class Analysis:
    """Analyse report."""
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def info(self):
        return self._info

    @property
    def name(self):
        return self._ptch.tuple[0]

    @property
    def alias(self):
        return self._acnm

    @property
    def protochain(self):
        return self._ptch

    ##########################################################################
    # Methods.
    ##########################################################################

    @staticmethod
    def analyse(file, length=None, *, _termination=False):
        """Analyse application layer packets."""
        if not _termination:
            # HTTP/1.* analysis
            flag, http = Analysis._analyse_httpv1(file, length)
            if flag:
                return Analysis(http.info, http.protochain, http.alias)

            # NOTE: due to format similarity of HTTP/2 and TLS/SSL, HTTP/2 won't be analysed before TLS/SSL is implemented.
            # NB: the NOTE abrove is deprecated, since validations are performed

            # HTTP/2 analysis
            flag, http = Analysis._analyse_httpv2(file, length)
            if flag:
                return Analysis(http.info, http.protochain, http.alias)

        # raw packet analysis
        raw = Raw(file, length)
        return Analysis(raw.info, raw.protochain, raw.alias)

    ##########################################################################
    # Data modules.
    ##########################################################################

    def __init__(self, info, ptch, acnm):
        self._info = info
        self._ptch = ptch
        self._acnm = acnm

    def __str__(self):
        return f'Analysis({self._ptch.alias[0]}, info={self._info})'

    def __repr__(self):
        return f'Analysis({self._ptch.alias[0]})'

    ##########################################################################
    # Utilities.
    ##########################################################################

    @staticmethod
    @seekset_ng
    def _analyse_httpv1(file, length):
        try:
            from jspcap.protocols.application.httpv1 import HTTPv1
            http = HTTPv1(file, length)
        except ProtocolError:
            return False, None
        return True, http

    @staticmethod
    @seekset_ng
    def _analyse_httpv2(file, length):
        try:
            from jspcap.protocols.application.httpv2 import HTTPv2
            http = HTTPv2(file, length)
        except ProtocolError:
            return False, None
        return True, http



# -*- coding: utf-8 -*-
"""user defined exceptions

`jspcap.exceptions` refined built-in exceptions. Make it
possible to show only user error stack infomation, when
exception raised on user's operation.

"""
import pathlib
import struct
import sys
import traceback


__all__ = [
    'BaseError',                                                # Exception
    'DigitError', 'IntError', 'RealError', 'ComplexError',      # TypeError
    'BoolError', 'BytesError', 'StringError', 'BytearrayError', # TypeError
    'DictError', 'ListError', 'TupleError', 'IterableError',    # TypeError
    'IOObjError', 'ProtocolUnbound', 'CallableError',           # TypeError
    'InfoError',                                                # TypeError
    'FormatError', 'UnsupportedCall',                           # AttributeError
    'FileError',                                                # IOError
    'FileExists',                                               # FileExistsError
    'FileNotFound',                                             # FileNotFoundError
    'ProtocolNotFound',                                         # IndexError
    'VersionError', 'IndexNotFound', 'ProtocolError',           # ValueError
    'ProtocolNotImplemented',                                   # NotImplementedError
    'StructError',                                              # struct.error
    'FragmentError', 'PacketError'                              # KeyError
]



def stacklevel():
    """Fetch curent stack level."""
    tb = traceback.extract_stack()
    for tbitem in tb:
        if pathlib.Path(tbitem[0]).match('*/jspcap/*'):
            index = tb.index(tbitem)
            break
    else:
        index = len(tb)
    return (index-1)


##############################################################################
# BaseError (abc of exceptions) session.
##############################################################################


class BaseError(Exception):
    """Base error class of all kinds.

    Cautions:
        * Turn off system-default traceback function by set `sys.tracebacklimit` to 0.
        * But bugs appear in Python 3.6, so we have to set `sys.tracebacklimit` to None.
            > this note is deprecated since Python fixed the problem above
        * In Python 2.7, `trace.print_stack(limit=None)` dose not support negative limit.

    """
    def __init__(self, message=None, *args, **kwargs):
        index = stacklevel()
        quiet = kwargs.pop('quiet', False)
        if not quiet and index:
            print('Traceback (most recent call last):')
            traceback.print_stack(limit=-index)

        sys.tracebacklimit = 0
        super().__init__(message, *args, **kwargs)


##############################################################################
# TypeError session.
##############################################################################


class DigitError(BaseError, TypeError):
    """The argument(s) must be (a) number(s)."""
    pass


class IntError(BaseError, TypeError):
    """The argument(s) must be integral."""
    pass


class RealError(BaseError, TypeError):
    """The function is not defined for real number."""
    pass


class ComplexError(BaseError, TypeError):
    """The function is not defined for complex instance."""
    pass


class BytesError(BaseError, TypeError):
    """The argument(s) must be bytes type."""
    pass


class BytearrayError(BaseError, TypeError):
    """The argument(s) must be bytearray type."""
    pass


class BoolError(BaseError, TypeError):
    """The argument(s) must be bool type."""
    pass


class StringError(BaseError, TypeError):
    """The argument(s) must be str type."""
    pass


class DictError(BaseError, TypeError):
    """The argument(s) must be dict type."""
    pass


class ListError(BaseError, TypeError):
    """The argument(s) must be list type."""
    pass


class TupleError(BaseError, TypeError):
    """The argument(s) must be tuple type."""
    pass


class IterableError(BaseError, TypeError):
    """The argument(s) must be iterable."""
    pass


class CallableError(BaseError, TypeError):
    """The argument(s) must be callable."""
    pass


class ProtocolUnbound(BaseError, TypeError):
    """Protocol slice unbound."""
    pass


class IOObjError(BaseError, TypeError):
    """The argument(s) must be file-like object."""
    pass


class InfoError(BaseError, TypeError):
    """The argument(s) must be Info instance."""
    pass


##############################################################################
# AttributeError session.
##############################################################################


class FormatError(BaseError, AttributeError):
    """Unknown format(s)."""
    pass


class UnsupportedCall(BaseError, AttributeError):
    """Unsupported function or property call."""
    pass


##############################################################################
# IOError session.
##############################################################################


class FileError(BaseError, IOError):
    """Wrong file format."""
    pass


##############################################################################
# FileExistsError session.
##############################################################################


class FileExists(BaseError, FileExistsError):
    """File already exists."""
    pass


##############################################################################
# FileNotFoundError session.
##############################################################################


class FileNotFound(BaseError, FileNotFoundError):
    """File not found."""
    pass


##############################################################################
# IndexError session.
##############################################################################


class ProtocolNotFound(BaseError, IndexError):
    """Protocol not found in ProtoChain."""
    pass


##############################################################################
# ValueError session.
##############################################################################


class VersionError(BaseError, ValueError):
    """Unknown IP version."""
    pass


class IndexNotFound(BaseError, ValueError):
    """Protocol not in ProtoChain."""
    pass


class ProtocolError(BaseError, ValueError):
    """Invalid protocol format."""
    pass

##############################################################################
# NotImplementedError session.
##############################################################################


class ProtocolNotImplemented(BaseError, NotImplementedError):
    """Protocol not implemented."""
    pass


##############################################################################
# struct.error session.
##############################################################################


class StructError(BaseError, struct.error):
    """Unpack failed."""
    pass


##############################################################################
# KeyError session.
##############################################################################


class FragmentError(BaseError, KeyError):
    """Invalid fragment dict."""
    pass


class PacketError(BaseError, KeyError):
    """Invalid packet dict."""
    pass



# -*- coding: utf-8 -*-
"""reassembly TCP datagram

`jspcap.reassembly.tcp` contains `TCP_Reassembly` only,
which reconstructs fragmented TCP packets back to origin.
The algorithm for TCP reassembly is decribed as below.

Notations:

    DSN     - Data Sequence Number 
    ACK     - TCP Acknowledgement
    FIN     - TCP Finish Flag
    SYN     - TCP Synchronisation Flag
    BUFID   - Buffer Identifier
    HDL     - Hole Discriptor List
    ISN     - Initial Sequence Number
    src     - source IP
    dst     - destionation IP
    srcport - source TCP port
    dstport - destionation TCP port

Algorithm:

    DO { 
        BUFID <- src|dst|srcport|dstport|ACK;  
        IF (SYN is true) { 
            IF (buffer with BUFID is allocated) { 
                flush all reassembly for this BUFID; 
                submit datagram to next step; 
            } 
        }  

        IF (no buffer with BUFID is allocated) { 
            allocate reassembly resources with BUFID; 
            ISN <- DSN; 
            put data from fragment into data buffer with BUFID
                [from octet fragment.first to octet fragment.last]; 
            update HDL; 
        }  

        IF (FIN is true) { 
            submit datagram to next step; 
            free all reassembly resources for this BUFID; 
            BREAK. 
        }
      } give up until (next fragment);

    update HDL: { 
        DO { 
            select the next hole descriptor from HDL;  

            IF (fragment.first >= hole.first) CONTINUE. 
            IF (fragment.last <= hole.first) CONTINUE.  

            delete the current entry from HDL;  

            IF (fragment.first >= hole.first) { 
                create new entry "new_hole" in HDL; 
                new_hole.first <- hole.first; 
                new_hole.last <- fragment.first - 1; 
                BREAK. 
            }  

            IF (fragment.last <= hole.last) { 
                create new entry "new_hole" in HDL; 
                new_hole.first <- fragment.last + 1 ;
                new_hole.last <- hole.last; 
                BREAK. 
            }
        } give up until (no entry from HDL)
     }

The following algorithm implementment is based on `IP Datagram
Reassembly Algorithm` introduced in RFC 815. It descripted an
algorithm dealing with `RCVBT` (fragment received bit table)
appeared in RFC 791. And here is the process:

1. Select the next hole descriptor from the hole descriptor
  list. If there are no more entries, go to step eight.
2. If fragment.first is greater than hole.last, go to step one.
3. If fragment.last is less than hole.first, go to step one.
4. Delete the current entry from the hole descriptor list.
5. If fragment.first is greater than hole.first, then create a
  new hole descriptor "new_hole" with new_hole.first equal to
  hole.first, and new_hole.last equal to fragment.first minus
  one.
6. If fragment.last is less than hole.last and
  fragment.more_fragments is true, then create a new hole descriptor
  "new_hole", with new_hole.first equal to fragment.last plus
  one and new_hole.last equal to hole.last.
7. Go to step one.
8. If the hole descriptor list is now empty, the datagram is now
  complete. Pass it on to the higher level protocol processor
  for further handling. Otherwise, return.

"""
import copy
import io
import sys

from jspcap.corekit.infoclass import Info
from jspcap.foundation.analysis import Analysis
from jspcap.reassembly.reassembly import Reassembly


__all__ = ['TCP_Reassembly']


class TCP_Reassembly(Reassembly):
    """Reassembly for TCP payload.

    Usage:
        >>> from jspcap.reassembly import TCP_Reassembly
        # Initialise instance:
        >>> tcp_reassembly = TCP_Reassembly()
        # Call reassembly:
        >>> tcp_reassembly(packet_dict)
        # Fetch result:
        >>> result = tcp_reassembly.datagram

    Properties:
        * name -- str, protocol of current packet
        * count -- int, total number of reassembled packets
        * datagram -- tuple, reassembled datagram, which structure may vary
                        according to its protocol
        * protocol -- str, protocol of current reassembly object

    Methods:
        * reassembly -- perform the reassembly procedure
        * submit -- submit reassembled payload
        * fetch -- fetch datagram
        * index -- return datagram index
        * run -- run automatically

    Attributes:
        * _strflg -- bool, stirct mode flag
        * _buffer -- dict, buffer field
        * _dtgram -- tuple, reassembled datagram

    Terminology:
        - packet_dict = Info(
            bufid = tuple(
                ip.src,                     # source IP address
                ip.dst,                     # destination IP address
                tcp.srcport,                # source port
                tcp.dstport,                # destination port
            ),
            num = frame.number,             # original packet range number
            ack = tcp.ack,                  # acknowledgement
            dsn = tcp.seq,                  # data sequence number
            syn = tcp.flags.syn,            # synchronise flag
            fin = tcp.flags.fin,            # finish flag
            len = tcp.raw_len,              # payload length, header excludes
            first = tcp.seq,                # this sequence number
            last = tcp.seq + tcp.raw_len,   # next (wanted) sequence number
            payload = tcp.raw,              # raw bytearray type payload
         )
        - (tuple) datagram
           |--> (Info) data
           |       |--> 'NotImplemented' : (bool) True --> implemented
           |       |--> 'id' : (Info) original packet identifier
           |       |                |--> 'src' --> (tuple)
           |       |                |                |--> (str) ip.src
           |       |                |                |--> (int) tcp.srcport
           |       |                |--> 'dst' --> (tuple)
           |       |                |                |--> (str) ip.dst
           |       |                |                |--> (int) tcp.dstport
           |       |                |--> 'ack' --> (int) original packet ACK number
           |       |--> 'index' : (tuple) packet numbers
           |       |                |--> (int) original packet range number
           |       |--> 'payload' : (bytes/None) reassembled application layer data
           |       |--> 'packets' : (tuple<Analysis>) analysed payload
           |--> (Info) data
           |       |--> 'NotImplemented' : (bool) False --> not implemented
           |       |--> 'id' : (Info) original packet identifier
           |       |                |--> 'src' --> (tuple)
           |       |                |                |--> (str) ip.src
           |       |                |                |--> (int) tcp.srcport
           |       |                |--> 'dst' --> (tuple)
           |       |                |                |--> (str) ip.dst
           |       |                |                |--> (int) tcp.dstport
           |       |                |--> 'ack' --> (int) original packet ACK number
           |       |--> 'ack' : (int) original packet ACK number
           |       |--> 'index' : (tuple) packet numbers
           |       |                |--> (int) original packet range number
           |       |--> 'payload' : (tuple/None) partially reassembled payload
           |       |                |--> (bytes/None) payload fragment
           |       |--> 'packets' : (tuple<Analysis>) analysed payloads
           |--> (Info) data ...
        - (dict) buffer --> memory buffer for reassembly
           |--> (tuple) BUFID : (dict)
           |       |--> ip.src      |
           |       |--> ip.dst      |
           |       |--> tcp.srcport |
           |       |--> tcp.dstport |
           |                        |--> 'hdl' : (list) hole descriptor list
           |                        |               |--> (Info) hole --> hole descriptor
           |                        |                       |--> "first" --> (int) start of hole
           |                        |                       |--> "last" --> (int) stop ofhole
           |                        |--> (int) ACK : (dict)
           |                        |               |--> 'ind' : (list) list of reassembled packets
           |                        |               |               |--> (int) packet range number
           |                        |               |--> 'isn' : (int) ISN of payload buffer
           |                        |               |--> 'len' : (int) length of payload buffer
           |                        |               |--> 'raw' : (bytearray) reassembled payload,
           |                        |                               holes set to b'\x00'
           |                        |--> (int) ACK ...
           |                        |--> ...
           |--> (tuple) BUFID ...

    """
    ##########################################################################
    # Methods.
    ##########################################################################

    @property
    def name(self):
        """Protocol of current packet."""
        return 'Transmission Control Protocol'

    @property
    def protocol(self):
        """Protocol of current reassembly object."""
        return 'TCP'

    ##########################################################################
    # Methods.
    ##########################################################################

    def reassembly(self, info):
        """Reassembly procedure.

        Positional arguments:
            * info -- Info, info dict of packets to be reassembled

        """
        BUFID = info.bufid  # Buffer Identifier
        DSN = info.dsn      # Data Sequence Number
        ACK = info.ack      # Acknowledgement Number
        FIN = info.fin      # Finish Flag (Termination)
        SYN = info.syn      # Synchronise Flag (Establishment)

        # when SYN is set, reset buffer of this session
        if SYN and BUFID in self._buffer:
            self._dtgram += self.submit(self._buffer[BUFID], bufid=BUFID)
            del self._buffer[BUFID]

        # initialise buffer with BUFID & ACK
        if BUFID not in self._buffer:
            self._buffer[BUFID] = {
                'hdl' : [Info(first=info.len, last=sys.maxsize),],
                ACK : dict(
                    ind = list(),
                    isn = info.dsn,
                    len = info.len,
                    raw = info.payload,
                ),
            }

        # initialise buffer with ACK
        if ACK not in self._buffer[BUFID]:
            self._buffer[BUFID][ACK] = dict(
                ind = list(),
                isn = info.dsn,
                len = info.len,
                raw = info.payload,
            )

        # append packet index
        self._buffer[BUFID][ACK]['ind'].append(info.num)

        # record fragment payload
        ISN = self._buffer[BUFID][ACK]['isn']   # Initial Sequence Number
        RAW = self._buffer[BUFID][ACK]['raw']   # Raw Payload Data
        if DSN >= ISN:  # if fragment goes after exsisting payload
            LEN = self._buffer[BUFID][ACK]['len']
            GAP = DSN - (ISN + LEN)     # gap length between payloads
            if GAP >= 0:    # if fragment goes after exsisting payload
                RAW += bytearray(GAP) + info.payload
            else:           # if fragment partially overlaps exsisting payload
                RAW[DSN-ISN:] = info.payload
        else:           # if fragment exceeds exsisting payload
            LEN = info.len
            GAP = ISN - (DSN + LEN)     # gap length between payloads
            self._buffer[BUFID][ACK]['isn'] = DSN
            if GAP >= 0:    # if fragment exceeds exsisting payload
                RAW = info.payload + bytearray(GAP) + RAW
            else:           # if fragment partially overlaps exsisting payload
                RAW = info.payload + RAW[ISN-GAP:]
        self._buffer[BUFID][ACK]['raw'] = RAW       # update payload datagram
        self._buffer[BUFID][ACK]['len'] = len(RAW)  # update payload length

        # update hole descriptor list
        HDL = copy.deepcopy(self._buffer[BUFID]['hdl'])
        for (index, hole) in enumerate(self._buffer[BUFID]['hdl']): # step one
            if info.first > hole.last:                              # step two
                continue
            if info.last < hole.first:                              # step three
                continue
            del HDL[index]                                          # step four
            if info.first > hole.first:                             # step five
                new_hole = Info(
                    first = hole.first,
                    last = info.first - 1,
                )
                HDL.insert(index, new_hole)
            if info.last < hole.last and not FIN:                   # step six
                new_hole = Info(
                    first = info.last + 1,
                    last = hole.last
                )
                HDL.insert(index+1, new_hole)
            break                                                   # step seven
        self._buffer[BUFID]['hdl'] = HDL                            # update HDL

        # when FIN is set, submit buffer of this session
        if FIN:
            self._dtgram += self.submit(self._buffer[BUFID], bufid=BUFID)
            del self._buffer[BUFID]

    def submit(self, buf, *, bufid):
        """Submit reassembled payload.

        Positional arguments:
            * buf -- dict, buffer dict of reassembled packets

        Keyword arguments:
            * bufid -- tuple, buffer identifier

        Returns:
            * list -- reassembled packets

        """
        datagram = []           # reassembled datagram
        HDL = buf.pop('hdl')    # hole descriptor list (remove from dict)

        # check through every buffer with ACK
        for (ack, buffer) in buf.items():
            # if this buffer is not implemented
            # go through every hole and extract received payload
            if len(HDL) > 2 and self._strflg:
                data = []
                start = stop = 0
                for hole in HDL:
                    stop = hole.first
                    byte = buffer['raw'][start:stop]
                    start = hole.last
                    if byte:    # strip empty payload
                        data.append(byte)
                byte = buffer['raw'][start:]
                if byte:    # strip empty payload
                    data.append(byte)
                if data:    # strip empty buffer
                    packet = Info(
                        NotImplemented = True,
                        id = Info(
                            src = (bufid[0], bufid[2]),
                            dst = (bufid[1], bufid[3]),
                            ack = ack,
                        ),
                        index = tuple(buffer['ind']),
                        payload = tuple(data) or None,
                        packets = tuple([ Analysis.analyse(io.BytesIO(frag), len(frag)) for frag in data ]),
                    )
                    datagram.append(packet)
            # if this buffer is implemented
            # export payload data & convert into bytes
            else:
                data = buffer['raw']
                if data:    # strip empty buffer
                    packet = Info(
                        NotImplemented = False,
                        id = Info(
                            src = (bufid[0], bufid[2]),
                            dst = (bufid[1], bufid[3]),
                            ack = ack,
                        ),
                        index = tuple(buffer['ind']),
                        payload = bytes(data) or None,
                        packets = (Analysis.analyse(io.BytesIO(data), len(data)),),
                    )
                    datagram.append(packet)
        return datagram



# -*- coding: utf-8 -*-
"""root application layer protocol

`jspcap.protocols.application.application` contains only
`Application`, which is a base class for application
layer protocols, eg. HTTP/1.*, HTTP/2 and etc.

"""
from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.protocol import Protocol
from jspcap.utilities.exceptions import UnsupportedCall


__all__ = ['Application']


class Application(Protocol):
    """Abstract base class for transport layer protocol family.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Application`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _make_protochain -- make ProtoChain instance for corresponding protocol

    """
    __layer__ = 'Application'

    ##########################################################################
    # Properties.
    ##########################################################################

    # protocol layer
    @property
    def layer(self):
        """Protocol layer."""
        return self.__layer__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _make_protochain(self):
        """Make ProtoChain instance for corresponding protocol."""
        self._protos = ProtoChain(self.__class__.__name__, None, self.alias)

    def _decode_next_layer(self, dict_, proto=None, length=None):
        """Deprecated."""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute '_decode_next_layer'")

    def _import_next_layer(self, proto, length):
        """Deprecated."""
        raise UnsupportedCall(f"'{self.__class__.__name__}' object has no attribute '_import_next_layer'")



# -*- coding: utf-8 -*-
"""extractor for PCAP files

`jspcap.tools.extraction` contains `Extractor` only, which
synthesises file I/O and protocol analysis, coordinates
information exchange in all network layers, extracst
parametres from a PCAP file.

"""
import importlib
import io
import os
import pathlib
import textwrap
import time
import traceback
import warnings

###############################################################################
# from jsformat import PLIST, JSON, Tree, JavaScript, XML
###############################################################################

from jspcap.corekit.infoclass import Info
from jspcap.protocols.pcap.frame import Frame
from jspcap.protocols.pcap.header import Header
from jspcap.utilities.exceptions import stacklevel, CallableError, \
        FileNotFound, UnsupportedCall, IterableError
from jspcap.utilities.warnings import FormatWarning, EngineWarning, \
        LayerWarning, ProtocolWarning, AttributeWarning

###############################################################################
# from jspcap.foundation.traceflow import TraceFlow
# from jspcap.reassembly.ipv4 import IPv4_Reassembly
# from jspcap.reassembly.ipv6 import IPv6_Reassembly
# from jspcap.reassembly.tcp import TCP_Reassembly
###############################################################################


__all__ = ['Extractor']


# check list
LAYER_LIST = {'None', 'Link', 'Internet', 'Transport', 'Application'}
PROTO_LIST = {
    'null', 'protocol', 'raw',                              # base protocols
    'header', 'frame',                                      # PCAP headers
    'link', 'arp', 'inarp', 'ethernet', 'l2tp', 'ospf', 'rarp', 'drarp', 'vlan',
                                                            # Link layer
    'internet', 'ah', 'hip', 'hopopt', 'ip', 'ipsec', 'ipv4', 'ipv6', 'ipv6_frag',
    'ipv6_opts', 'ipv6_route', 'ipx', 'mh',                 # Internet layer
    'transport', 'tcp', 'udp',                              # Transport layer
    'application', 'http', 'httpv1', 'httpv2',              # Application layer
}


class Extractor:
    """Extractor for PCAP files.

    Properties:
        * info -- VerionInfo, version of input PCAP file
        * length -- int, frame number (of current extracted frame or all)
        * format -- str, format of output file
        * input -- str, name of input PCAP file
        * output -- str, name of output file
        * header -- Info, global header
        * frames -- tuple<Info>, extracted frames
        * protocol -- ProtoChain, protocol chain of current/last frame
        * reassembly -- Info, frame record for reassembly
            |--> tcp -- tuple<TCP_Reassembly>, TCP payload fragment reassembly
            |--> ipv4 -- tuple<IPv4_Reassembly>, IPv4 frame fragment reassembly
            |--> ipv6 -- tuple<IPv6_Reassembly>, IPv6 frame fragment reassembly

    Methods:
        * make_name -- formatting input & output file name
        * record_header -- extract global header
        * record_frames -- extract frames

    Attributes:
        * _flag_a -- bool, if run automatically to the end

        * _ifnm -- str, input file name (aka _ifile.name)
        * _ofnm -- str, output file name (aka _ofile.name)
        * _type -- str, output file kind (aka _ofile.kind)
        * _fext -- str, output file extension

        * _ifile -- FileIO, input file object
        * _ofile -- object, temperory output writer

        * _frnum -- int, frame number
        * _frame -- list, each item contains `Info` of a record/package
            |--> gbhdr -- Info object, global header
            |--> frame 1 -- Info object, record/package header
            |       |--> Info object, first (link layer) header
            |       |--> Info object, next protocol header
            |       |--> ......
            |       |--> Info object, optional protocol trailer
            |--> frame 2 -- Info object, record/package header
            |       |--> ......
        * _reasm -- list, frame reassembly instance
            |--> IPv4 -- IPv4_Reassembly, reassembly instance
            |--> IPv6 -- IPv6_Reassembly, reassembly instance
            |--> TCP -- TCP_Reassembly, reassembly instance

        * _gbhdr -- Info object, the global header
        * _dlink -- str, data link layer protocol of input file
        * _vinfo -- VersionInfo, version of input file
        * _proto -- str, protocol chain of current frame

        * _ip -- bool, flag if perform IPv4 & IPv6 reassembly
        * _ipv4 -- bool, flag if perform IPv4 reassembly
        * _ipv6 -- bool, flag if perform IPv6 reassembly
        * _tcp -- bool, flag if perform TCP payload reassembly

    Utilities:
        * _read_frame -- read frames
        * _tcp_reassembly -- store data for TCP reassembly
        * _ipv4_reassembly -- store data for IPv4 reassembly
        * _ipv6_reassembly -- store data for IPv6 reassembly

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def info(self):
        if self._exeng in ('scapy',):
            raise UnsupportedCall(f"'Extractor(engine={self._exeng})' object has no attribute 'info'")
        return self._vinfo

    @property
    def length(self):
        return self._frnum

    @property
    def format(self):
        if self._flag_q:
            raise UnsupportedCall("'Extractor(nofile=True)' object has no attribute 'format'")
        return self._type

    @property
    def input(self):
        return self._ifnm

    @property
    def output(self):
        if self._flag_q:
            raise UnsupportedCall("'Extractor(nofile=True)' object has no attribute 'format'")
        return self._ofnm

    @property
    def header(self):
        if self._exeng in ('scapy',):
            raise UnsupportedCall(f"'Extractor(engine={self._exeng})' object has no attribute 'header'")
        return self._gbhdr

    @property
    def protocol(self):
        if self._flag_a:
            raise UnsupportedCall(f"'Extractor(auto=True)' object has no attribute 'protocol'")
        return self._proto

    @property
    def frame(self):
        if self._flag_d:
            return tuple(self._frame)
        raise UnsupportedCall("'Extractor(store=False)' object has no attribute 'frame'")

    @property
    def reassembly(self):
        data = Info(
            ipv4 = tuple(self._reasm[0].datagram) if self._ipv4 else None,
            ipv6 = tuple(self._reasm[1].datagram) if self._ipv6 else None,
            tcp = tuple(self._reasm[2].datagram) if self._tcp else None,
        )
        return data

    @property
    def trace(self):
        if self._flag_t:
            return self._trace.index
        raise UnsupportedCall("'Extractor(trace=False)' object has no attribute 'trace'")

    @property
    def engine(self):
        return self._exeng

    ##########################################################################
    # Methods.
    ##########################################################################

    def run(self):
        """Start extraction."""
        if self._exeng == 'dpkt':
            flag, engine = self.import_test(self._exeng, name='DPKT')
            if flag:    pass
        elif self._exeng == 'scapy':
            flag, engine = self.import_test('scapy.all', name='Scapy')
            if flag:    return self._run_scapy(engine)
        elif self._exeng == 'pyshark':
            flag, engine = self.import_test(self._exeng, name='PyShark')
            if flag:    pass
        elif self._exeng not in ('default', 'jspcap'):
            warnings.warn(f'unsupported extraction engine: {self._exeng}; '
                            'using default engine instead',
                            EngineWarning, stacklevel=stacklevel())

        # using default/jspcap engine
        self.record_header()            # read PCAP global header
        self.record_frames()            # read frames

    @staticmethod
    def import_test(engine, *, name=None):
        try:
            engine = importlib.import_module(engine)
            return True, engine
        except ImportError:
            warnings.warn(f'extraction engine {name or engine} not installed; '
                            'using default engine instead',
                            EngineWarning, stacklevel=stacklevel())
        return False, None

    def check_list(self):
        layer = self._exlyr
        protocol = self._exptl
        if self._exeng in ('scapy',) and (layer or protocol):
            warnings.warn(f"'Extractor(engine={self._exeng})' does not support protocol and layer threshold; "
                            f"'layer={layer}' and 'protocol={protocol}' ignored", AttributeWarning, stacklevel=stacklevel())

        if layer is not None:
            if layer.capitalize() not in LAYER_LIST:
                warnings.warn(f'unrecognised layer: {layer}',
                                LayerWarning, stacklevel=stacklevel())
        if protocol is not None:
            def check_protocol(*args):
                for arg in args:
                    if arg.lower() not in PROTO_LIST:
                        warnings.warn(f'unrecognised protocol: {protocol}',
                                        ProtocolWarning, stacklevel=stacklevel())
            if isinstance(protocol, tuple): check_protocol(*protocol)
            else:                           check_protocol(protocol)

    @classmethod
    def make_name(cls, fin, fout, fmt, extension, *, files=False, nofile=False):
        if fin is None:
            ifnm = 'in.pcap'
        else:
            if extension:
                ifnm = fin if os.path.splitext(fin)[1] == '.pcap' else f'{fin}.pcap'
            else:
                ifnm = fin

        if not os.path.isfile(ifnm):
            raise FileNotFound(f"[Errno 2] No such file or directory: '{ifnm}'")

        if nofile:
            ofnm = None
            ext = None
        else:
            fmt_none = (fmt is None)

            if fmt == 'html':   ext = 'js'
            elif fmt == 'tree': ext = 'txt'
            else:               ext = fmt

            if fout is None:
                if fmt_none:
                    raise FormatError('Output format unspecified.')
                elif files:
                    ofnm = 'out'
                    pathlib.Path(ofnm).mkdir(parents=True, exist_ok=True)
                else:
                    ofnm = f'out.{ext}'
            else:
                name, fext = os.path.splitext(fout)
                pathlib.Path(fout).parent.mkdir(parents=True, exist_ok=True)
                if fext:
                    files = False
                    ofnm = fout
                    fmt = fmt or fext[1:] or None
                    if fmt is None:
                        raise FormatError('Output format unspecified.')
                elif fmt_none:
                    raise FormatError('Output format unspecified.')
                elif files:
                    ofnm = fout
                    pathlib.Path(ofnm).mkdir(parents=True, exist_ok=True)
                elif extension: ofnm = f'{fout}.{ext}'
                else:           ofnm = fout

        return ifnm, ofnm, fmt, ext, files

    def record_header(self):
        """Read global header.

        - Extract global header.
        - Make Info object out of header properties.
        - Append Info.
        - Write plist file.

        """
        self._gbhdr = Header(self._ifile)
        self._dlink = self._gbhdr.protocol
        self._vinfo = self._gbhdr.version
        if not self._flag_q:
            if self._flag_f:
                ofile = self._ofile(f'{self._ofnm}/Global Header.{self._fext}')
                ofile(self._gbhdr.info, name='Global Header')
                self._type = ofile.kind
            else:
                self._ofile(self._gbhdr.info, name='Global Header')
                self._type = self._ofile.kind

    def record_frames(self):
        if self._flag_a:
            while True:
                try:
                    self._read_frame()
                except EOFError:
                    # quit when EOF
                    break
            self._ifile.close()

    ##########################################################################
    # Data modules.
    ##########################################################################

    # Not hashable
    __hash__ = None

    def __init__(self, *,
                    fin=None, fout=None, format=None,                           # basic settings
                    auto=True, extension=True, store=True,                      # internal settings
                    files=False, nofile=False, verbose=False,                   # output settings
                    engine=None, layer=None, protocol=None,                     # extraction settings
                    ip=False, ipv4=False, ipv6=False, tcp=False, strict=False,  # reassembly settings
                    trace=False, trace_fout=None, trace_format=None):           # trace settings
        """Initialise PCAP Reader.

        Keyword arguments:
            * fin  -- str, file name to be read; if file not exist, raise an error
            * fout -- str, file name to be written
            * format  -- str, file format of output
                            <keyword> 'plist' / 'json' / 'tree' / 'html'

            * auto -- bool, if automatically run till EOF (default is True)
                            <keyword> True / False
            * extension -- bool, if check and append axtensions to output file (default is True)
                            <keyword> True / False
            * store -- bool, if store extracted packet info (default is True)
                            <keyword> True / False

            * files -- bool, if split each frame into different files (default is False)
                            <keyword> True / False
            * nofile -- bool, if no output file is to be dumped (default is False)
                            <keyword> True / False
            * verbose -- bool, if print verbose output information (default is False)
                            <keyword> True / False

            * engine -- str, extraction engine to be used
                            <keyword> 'default | jspcap'
            * layer -- str, extract til which layer
                            <keyword> 'Link' / 'Internet' / 'Transport' / 'Application'
            * protocol -- str, extract til which protocol
                            <keyword> available protocol name

            * ip -- bool, if record data for IPv4 & IPv6 reassembly (default is False)
                            <keyword> True / False
            * ipv4 -- bool, if perform IPv4 reassembly (default is False)
                            <keyword> True / False
            * ipv6 -- bool, if perform IPv6 reassembly (default is False)
                            <keyword> True / False
            * tcp -- bool, if perform TCP reassembly (default is False)
                            <keyword> True / False
            * strict -- bool, if set strict flag for reassembly (default is False)
                            <keyword> True / False

            * trace -- bool, if trace TCP traffic flows (default is False)
                            <keyword> True / False
            * trace_fout -- str, path name for flow tracer if necessary
            * trace_format -- str, output file format of flow tracer
                            <keyword> 'plist' / 'json' / 'tree' / 'html' / 'pcap'

        """
        ifnm, ofnm, fmt, ext, files = \
            self.make_name(fin, fout, format, extension, files=files, nofile=nofile)

        self._ifnm = ifnm               # input file name
        self._ofnm = ofnm               # output file name
        self._fext = ext                # output file extension

        self._flag_a = auto             # auto extract flag
        self._flag_d = store            # store data flag
        self._flag_f = files            # split file flag
        self._flag_q = nofile           # no output flag
        self._flag_t = trace            # trace flag
        self._flag_v = verbose          # verbose output flag

        self._frnum = 0                 # frame number
        self._frame = list()            # frame record
        self._proto = None              # frame ProtoChain

        self._reasm = [None] * 3        # frame record for reassembly (IPv4 / IPv6 / TCP)
        self._trace = NotImplemented    # flow tracer

        self._ipv4 = ipv4 or ip         # IPv4 Reassembly
        self._ipv6 = ipv6 or ip         # IPv6 Reassembly
        self._tcp = tcp                 # TCP Reassembly

        self._exlyr = layer or 'none'                       # extract til layer
        self._exptl = protocol or 'null'                    # extract til protocol
        self._exeng = (engine or 'default').lower()         # extract using engine
        self.check_list()                                   # check layer & protocol

        if self._ipv4:
            from jspcap.reassembly.ipv4 import IPv4_Reassembly
            self._reasm[0] = IPv4_Reassembly(strict=strict)
        if self._ipv6:
            from jspcap.reassembly.ipv6 import IPv6_Reassembly
            self._reasm[1] = IPv6_Reassembly(strict=strict)
        if self._tcp:
            from jspcap.reassembly.tcp import TCP_Reassembly
            self._reasm[2] = TCP_Reassembly(strict=strict)

        if trace:
            from jspcap.foundation.traceflow import TraceFlow
            self._trace = TraceFlow(fout=trace_fout, format=trace_format)

        self._ifile = open(ifnm, 'rb')                      # input file
        if not self._flag_q:
            if fmt == 'plist':
                from jsformat import PLIST as output        # output PLIST file
            elif fmt == 'json':
                from jsformat import JSON as output         # output JSON file
            elif fmt == 'tree':
                from jsformat import Tree as output         # output treeview text file
            elif fmt == 'html':
                from jsformat import JavaScript as output   # output JavaScript file
            elif fmt == 'xml':
                from jsformat import XML as output          # output XML file
            else:
                from jspcap.dumpkit import NotImplementedIO as output
                                                            # no output file
                warnings.warn(f'unsupported output format: {fmt}; '
                                'disabled file output feature',
                                FormatWarning, stacklevel=stacklevel())
            self._ofile = output if self._flag_f else output(ofnm)
                                                            # output file
        self.run()                      # start extraction

    def __iter__(self):
        if not self._flag_a:
            return self
        raise IterableError("'Extractor(auto=True)' object is not iterable")

    def __next__(self):
        try:
            return self._read_frame()
        except EOFError:
            self._ifile.close()
            raise StopIteration

    def __call__(self):
        if not self._flag_a:
            try:
                return self._read_frame()
            except EOFError as error:
                self._ifile.close()
                raise error from None
        raise CallableError("'Extractor(auto=True)' object is not callable")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._ifile.close()

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_frame(self):
        """Read frames.

        - Extract frames and each layer of packets.
        - Make Info object out of frame properties.
        - Append Info.
        - Write plist & append Info.

        """
        # read frame header
        frame = Frame(self._ifile, num=self._frnum+1, proto=self._dlink,
                        layer=self._exlyr, protocol=self._exptl)
        if self._flag_v:
            print(f' - Frame {self._frnum:>3d}: {frame.protochain}')

        # write plist
        frnum = f'Frame {self._frnum+1}'
        if not self._flag_q:
            if self._flag_f:
                ofile = self._ofile(f'{self._ofnm}/{frnum}.{self._fext}')
                ofile(frame.info, name=frnum)
            else:
                self._ofile(frame.info, name=frnum)

        # record frames
        self._frnum += 1
        self._proto = frame.protochain.chain
        if self._flag_d:
            self._frame.append(frame)

        # record fragments
        if self._tcp:
            self._tcp_reassembly(frame)
        if self._ipv4:
            self._ipv4_reassembly(frame)
        if self._ipv6:
            self._ipv6_reassembly(frame)

        # trace flows
        if self._flag_t:
            self._tcp_traceflow(frame)

        # return frame record
        return frame

    def _ipv4_reassembly(self, frame):
        """Store data for IPv4 reassembly."""
        if 'IPv4' in frame:
            ipv4 = frame['IPv4']
            if ipv4.flags.df:   return                  # dismiss not fragmented frame
            data = dict(
                bufid = (
                    ipv4.src,                           # source IP address
                    ipv4.dst,                           # destination IP address
                    ipv4.id,                            # identification
                    ipv4.proto,                         # payload protocol type
                ),
                num = frame.info.number,                # original packet range number
                fo = ipv4.frag_offset,                  # fragment offset
                ihl = ipv4.hdr_len,                     # internet header length
                mf = ipv4.flags.mf,                     # more fragment flag
                tl = ipv4.len,                          # total length, header includes
                header = bytearray(ipv4.packet.header), # raw bytearray type header
                payload = bytearray(ipv4.packet.payload or b''),
                                                        # raw bytearray type payload
            )
            self._reasm[0](data)

    def _ipv6_reassembly(self, frame):
        """Store data for IPv6 reassembly."""
        if 'IPv6' in frame:
            ipv6 = frame['IPv6']
            if 'frag' not in ipv6:  return              # dismiss not fragmented frame
            data = dict(
                bufid = (
                    ipv6.src,                           # source IP address
                    ipv6.dst,                           # destination IP address
                    ipv6.label,                         # label
                    ipv6.ipv6_frag.next,                # next header field in IPv6 Fragment Header
                ),
                num = frame.info.number,                # original packet range number
                fo = ipv6.ipv6_frag.offset,             # fragment offset
                ihl = ipv6.hdr_len,                     # header length, only headers before IPv6-Frag
                mf = ipv6.ipv6_frag.mf,                 # more fragment flag
                tl = ipv6.hdr_len + ipv6.raw_len,       # total length, header includes
                header = bytearray(ipv6.fragment.header),
                                                        # raw bytearray type header before IPv6-Frag
                payload = bytearray(ipv6.fragment.payload or b''),
                                                        # raw bytearray type payload after IPv6-Frag
            )
            self._reasm[1](data)

    def _tcp_reassembly(self, frame):
        """Store data for TCP reassembly."""
        if 'TCP' in frame:
            ip = frame['IPv4'] if 'IPv4' in frame else frame['IPv6']
            tcp = frame['TCP']
            data = dict(
                bufid = (
                    ip.src,                             # source IP address
                    ip.dst,                             # destination IP address
                    tcp.srcport,                        # source port
                    tcp.dstport,                        # destination port
                ),
                num = frame.info.number,                # original packet range number
                ack = tcp.ack,                          # acknowledgement
                dsn = tcp.seq,                          # data sequence number
                syn = tcp.flags.syn,                    # synchronise flag
                fin = tcp.flags.fin,                    # finish flag
                payload = bytearray(tcp.packet.payload or b''),
                                                        # raw bytearray type payload
            )
            raw_len = len(data['payload'])              # payload length, header excludes
            data['first'] = tcp.seq                     # this sequence number
            data['last'] = tcp.seq + raw_len            # next (wanted) sequence number
            data['len'] = raw_len                       # payload length, header excludes
            self._reasm[2](data)

    def _tcp_traceflow(self, frame):
        """Trace packet flow for TCP."""
        if 'TCP' in frame:
            ip = frame['IPv4'] if 'IPv4' in frame else frame['IPv6']
            tcp = frame['TCP']
            data = dict(
                protocol = self._dlink,                 # data link type from global header
                index = frame.info.number,              # frame number
                frame = frame.info,                     # extracted frame info
                syn = tcp.flags.syn,                    # TCP synchronise (SYN) flag
                fin = tcp.flags.fin,                    # TCP finish (FIN) flag
                src = ip.src,                           # source IP
                dst = ip.dst,                           # destination IP
                srcport = tcp.srcport,                  # TCP source port
                dstport = tcp.dstport,                  # TCP destination port
                timestamp = frame.info.time_epoch,      # frame timestamp
            )
            self._trace(data)

    def _run_scapy(self, scapy_all):
        """Call scapy.all.sniff to extract PCAP files."""
        if not self._flag_a:
            self._flag_a = True
            warnings.warn(f"'Extractor(engine=scapy)' object is not iterable; "
                            "so 'auto=False' will be ignored", AttributeWarning, stacklevel=stacklevel())

        def _scapy_packet2dict(packet):
            """Convert Scapy packet into dict."""
            dict_ = packet.fields
            payload = packet.payload
            if not isinstance(payload, scapy_all.packet.NoPayload):
                dict_[payload.name] = _scapy_packet2dict(payload)
            return dict_

        def _scapy_tcp_reassembly(packet, *, count=NotImplemented):
            """Store data for IPv4 reassembly."""
            if 'IP' in packet:
                ipv4 = packet['IP']
                if ipv4.flags.DF:   return                  # dismiss not fragmented packet
                data = dict(
                    bufid = (
                        ipv4.src,                           # source IP address
                        ipv4.dst,                           # destination IP address
                        ipv4.id,                            # identification
                        ipv4.proto,                         # payload protocol type
                    ),
                    num = count,                            # original packet range number
                    fo = ipv4.frag,                         # fragment offset
                    ihl = ipv4.ihl,                         # internet header length
                    mf = ipv4.flags.MF,                     # more fragment flag
                    tl = ipv4.len,                          # total length, header includes
                    header = bytearray(ipv4.raw_packet_cache),
                                                            # raw bytearray type header
                    payload = bytearray(bytes(ipv4.payload)),
                                                            # raw bytearray type payload
                )
                self._reasm[0](data)

        def _scapy_ipv6_reassembly(packet, *, count=NotImplemented):
            """Store data for IPv6 reassembly."""
            if 'IPv6' in packet:
                ipv6 = packet['IPv6']
                if scapy_all.IPv6ExtHdrFragment not in ipv6:
                    return                                  # dismiss not fragmented packet
                data = dict(
                    bufid = (
                        ipv6.src,                           # source IP address
                        ipv6.dst,                           # destination IP address
                        ipv6.fl,                            # label
                        ipv6['IPv6ExtHdrFragment'].nh,      # next header field in IPv6 Fragment Header
                    ),
                    num = count,                            # original packet range number
                    fo = ipv6['IPv6ExtHdrFragment'].offset, # fragment offset
                    ihl = len(ipv6) - len(ipv6['IPv6ExtHdrFragment']),
                                                            # header length, only headers before IPv6-Frag
                    mf = ipv6['IPv6ExtHdrFragment'].m,      # more fragment flag
                    tl = len(ipv6),                         # total length, header includes
                    header = bytearray(bytes(ipv6)[:-len(ipv6['IPv6ExtHdrFragment'])]),
                                                            # raw bytearray type header before IPv6-Frag
                    payload = bytearray(bytes(ipv6['IPv6ExtHdrFragment'].payload)),
                                                            # raw bytearray type payload after IPv6-Frag
                )
                self._reasm[1](data)

        def _scapy_tcp_reassembly(packet, *, count=NotImplemented):
            """Store data for TCP reassembly."""
            if 'TCP' in packet:
                ip = packet['IP'] if 'IP' in packet else packet['IPv6']
                tcp = packet['TCP']
                data = dict(
                    bufid = (
                        ip.src,                             # source IP address
                        ip.dst,                             # destination IP address
                        tcp.sport,                          # source port
                        tcp.dport,                          # destination port
                    ),
                    num = count,                            # original packet range number
                    ack = tcp.ack,                          # acknowledgement
                    dsn = tcp.seq,                          # data sequence number
                    syn = tcp.flags.S,                      # synchronise flag
                    fin = tcp.flags.F,                      # finish flag
                    payload = bytearray(bytes(tcp.payload)),# raw bytearray type payload
                )
                raw_len = len(tcp.payload)                  # payload length, header excludes
                data['first'] = tcp.seq                     # this sequence number
                data['last'] = tcp.seq + raw_len            # next (wanted) sequence number
                data['len'] = raw_len                       # payload length, header excludes
                self._reasm[2](data)

        def _scapy_tcp_traceflow(packet, *, count=NotImplemented):
            """Trace packet flow for TCP."""
            if 'TCP' in packet:
                ip = packet['IP'] if 'IP' in packet else packet['IPv6']
                tcp = packet['TCP']
                data = dict(
                    protocol = packet.name,                 # data link type from global header
                    index = count,                          # frame number
                    frame = _scapy_packet2dict(packet),     # extracted packet
                    syn = tcp.flags.S,                      # TCP synchronise (SYN) flag
                    fin = tcp.flags.F,                      # TCP finish (FIN) flag
                    src = ip.src,                           # source IP
                    dst = ip.dst,                           # destination IP
                    srcport = tcp.sport,                    # TCP source port
                    dstport = tcp.dport,                    # TCP destination port
                    timestamp = time.time(),                # timestamp
                )
                self._trace(data)

        # extract & analyse file
        sniffed = scapy_all.sniff(offline=self._ifnm)
        for packet in sniffed:
            self._frnum += 1
            if self._flag_v:
                print(f' - Frame {self._frnum:>3d}: {packet.summary()}')

            # write plist
            frnum = f'Frame {self._frnum}'
            if not self._flag_q:
                info = {packet.name: _scapy_packet2dict(packet)}
                if self._flag_f:
                    ofile = self._ofile(f'{self._ofnm}/{frnum}.{self._fext}')
                    ofile(info, name=frnum)
                else:
                    self._ofile(info, name=frnum)

            # record frames
            if self._flag_d:
                self._frame.append(packet)

            # record fragments
            if self._tcp:
                _scapy_tcp_reassembly(packet, count=self._frnum)
            if self._ipv4:
                _scapy_ipv4_reassembly(packet, count=self._frnum)
            if self._ipv6:
                _scapy_ipv6_reassembly(packet, count=self._frnum)

            # trace flows
            if self._flag_t:
                _scapy_tcp_traceflow(packet)

        # aftermath
        self._ifile.close()



# -*- coding: utf-8 -*-
"""link layer protocols

`jspcap.protocols.link` is collection of all protocols in
link layer, with detailed implementation and methods.

"""
# TODO: Implements DSL, EAPOL, FDDI, ISDN, NDP, PPP.

# Base Class for Link Layer
from jspcap.protocols.link.link import Link

# Utility Classes for Protocols
from jspcap.protocols.link.arp import ARP
from jspcap.protocols.link.arp import ARP as InARP
from jspcap.protocols.link.ethernet import Ethernet
from jspcap.protocols.link.l2tp import L2TP
from jspcap.protocols.link.ospf import OSPF
from jspcap.protocols.link.rarp import RARP
from jspcap.protocols.link.rarp import RARP as DRARP
from jspcap.protocols.link.vlan import VLAN

# Link-Layer Header Type Values
from jspcap.protocols.link.link import LINKTYPE


__all__ = [
    'LINKTYPE',                             # Protocol Numbers
    'ARP', 'DRARP', 'Ethernet', 'InARP', 'L2TP', 'OSPF', 'RARP', 'VLAN',
                                            # Link Layer Protocols
]



"""hypertext transfer protocol (HTTP/1.*)

`jspcap.protocols.application.httpv1` contains `HTTPv1`
only, which implements extractor for Hypertext Transfer
Protocol (HTTP/1.*), whose structure is described as
below.

METHOD URL HTTP/VERSION\r\n :==: REQUEST LINE
<key> : <value>\r\n         :==: REQUEST HEADER
............  (Elipsis)     :==: REQUEST HEADER
\r\n                        :==: REQUEST SEPERATOR
<body>                      :==: REQUEST BODY (optional)

HTTP/VERSION CODE DESP \r\n :==: RESPONSE LINE
<key> : <value>\r\n         :==: RESPONSE HEADER
............  (Elipsis)     :==: RESPONSE HEADER
\r\n                        :==: RESPONSE SEPERATOR
<body>                      :==: RESPONSE BODY (optional)

"""
import re

import chardet

from jspcap.corekit.infoclass import Info
from jspcap.protocols.application.http import HTTP
from jspcap.utilities.exceptions import ProtocolError


__all__ = ['HTTPv1']


# utility regular expressions
_RE_METHOD = re.compile(rb'GET|HEAD|POST|PUT|DELETE|CONNECT|OPTIONS|TRACE')
_RE_VERSION = re.compile(rb'HTTP/(?P<version>\d\.\d)')
_RE_STATUS = re.compile(rb'\d{3}')


class HTTPv1(HTTP):
    """This class implements Hypertext Transfer Protocol (HTTP/1.*).

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Application`
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_http -- read Hypertext Transfer Protocol (HTTP/1.*)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _make_protochain -- make ProtoChain instance for corresponding protocol
        * _http_decode -- test and decode HTTP parameters
        * _read_http_header -- read HTTP/1.* header
        * _read_http_body -- read HTTP/1.* body

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def alias(self):
        """Acronym of current protocol."""
        return f'HTTP/{self._info.header[self._info.receipt].version}'

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_http(self, length):
        """Read Hypertext Transfer Protocol (HTTP/1.*).

        Structure of HTTP/1.* packet [RFC 7230]:
            HTTP-message    :==:    start-line
                                    *( header-field CRLF )
                                    CRLF
                                    [ message-body ]

        """
        if length is None:
            length =  len(self)

        packet = self._file.read(length)
        try:
            header, body = packet.split(b'\r\n\r\n', 1)
        except ValueError:
            raise ProtocolError('HTTP: invalid format', quiet=True)

        header_unpacked, http_receipt = self._read_http_header(header)
        body_unpacked = self._read_http_body(body) or None

        http = dict(
            receipt = http_receipt,
            header = header_unpacked,
            body = body_unpacked,
            raw = dict(
                header = header,
                body = body,
                packet = self._read_packet(length),
            ),
        )

        return http

    ##########################################################################
    # Data models.
    ##########################################################################

    def __length_hint__(self):
        pass

    @classmethod
    def __index__(cls):
        return cls.__name__

    ##########################################################################
    # Utilities.
    ##########################################################################

    @staticmethod
    def _http_decode(byte):
        charset = chardet.detect(byte)['encoding']
        if charset:
            try:
                return byte.decode(charset)
            except Exception:
                pass
        return byte

    def _read_http_header(self, header):
        """Read HTTP/1.* header.

        Structure of HTTP/1.* header [RFC 7230]:
            start-line      :==:    request-line / status-line
            request-line    :==:    method SP request-target SP HTTP-version CRLF
            status-line     :==:    HTTP-version SP status-code SP reason-phrase CRLF
            header-field    :==:    field-name ":" OWS field-value OWS

        """
        try:
            startline, headerfield = header.split(b'\r\n', 1)
            para1, para2, para3 = re.split(b'\s+', startline, 2)
            fields = headerfield.split(b'\r\n')
            lists = ( re.split(b'\s*:\s*', field, 1) for field in fields )
        except ValueError:
            raise ProtocolError('HTTP: invalid format', quiet=True)

        match1 = re.match(_RE_METHOD, para1)
        match2 = re.match(_RE_VERSION, para3)
        match3 = re.match(_RE_VERSION, para1)
        match4 = re.match(_RE_STATUS, para2)
        if match1 and match2:
            receipt = 'request'
            header = dict(
                request = dict(
                    method = self._http_decode(para1),
                    target = self._http_decode(para2),
                    version = self._http_decode(match2.group('version')),
                ),
            )
        elif match3 and match4:
            receipt = 'response'
            header = dict(
                response = dict(
                    version = self._http_decode(match3.group('version')),
                    status = int(para2),
                    phrase = self._http_decode(para3),
                ),
            )
        else:
            raise ProtocolError('HTTP: invalid format', quiet=True)

        try:
            for item in lists:
                key = self._http_decode(item[0].strip().replace(b'request', b'request_field').replace(b'response', b'response_field'))
                value = self._http_decode(item[1].strip())
                if key in header:
                    if isinstance(header[key], tuple):
                        header[key] += (value,)
                    else:
                        header[key] = (header[key], value)
                else:
                    header[key] = value
        except IndexError:
            raise ProtocolError('HTTP: invalid format', quiet=True)

        return header, receipt

    def _read_http_body(self, body):
        """Read HTTP/1.* body."""
        return self._http_decode(body)



# -*- coding: utf-8 -*-
"""user datagram protocol

`jspcap.protocols.transport.udp` contains `UDP` only,
which implements extractor for User Datagram Protocol
(UDP), whose structure is described as below.

 0      7 8     15 16    23 24    31
+--------+--------+--------+--------+
|     Source      |   Destination   |
|      Port       |      Port       |
+--------+--------+--------+--------+
|                 |                 |
|     Length      |    Checksum     |
+--------+--------+--------+--------+
|
|          data octets ...
+---------------- ...

"""
from jspcap.corekit.infoclass import Info
from jspcap.protocols.transport.transport import Transport


__all__ = ['UDP']


class UDP(Transport):
    """This class implements User Datagram Protocol.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Transport`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance
        * src -- int, source port
        * dst -- int, destination port

    Methods:
        * read_udp -- read User Datagram Protocol (UDP)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'User Datagram Protocol'

    @property
    def length(self):
        """Header length of current protocol."""
        return 8

    @property
    def src(self):
        """Source port."""
        return self._info.src

    @property
    def dst(self):
        """Destination port."""
        return self._info.dst

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_udp(self, length):
        """Read User Datagram Protocol (UDP).

        Structure of UDP header [RFC 768]:
             0      7 8     15 16    23 24    31
            +--------+--------+--------+--------+
            |     Source      |   Destination   |
            |      Port       |      Port       |
            +--------+--------+--------+--------+
            |                 |                 |
            |     Length      |    Checksum     |
            +--------+--------+--------+--------+
            |
            |          data octets ...
            +---------------- ...

            Octets      Bits        Name                    Discription
              0           0     udp.srcport             Source Port
              2          16     udp.dstport             Destination Port
              4          32     udp.len                 Length (header includes)
              6          48     udp.checksum            Checksum

        """
        if length is None:
            length = len(self)

        _srcp = self._read_unpack(2)
        _dstp = self._read_unpack(2)
        _tlen = self._read_unpack(2)
        _csum = self._read_fileng(2)

        udp = dict(
            srcport = _srcp,
            dstport = _dstp,
            len = _tlen,
            checksum = _csum,
        )

        length = udp['len'] - 8
        udp['packet'] = self._read_packet(header=8, payload=length)

        return self._decode_next_layer(udp, None, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, **kwargs):
        self._file = _file
        self._info = Info(self.read_udp(length))

    def __length_hint__(self):
        return 8



"""hypertext transfer protocol (HTTP/2)

`jspcap.protocols.application.httpv2` contains `HTTPv2`
only, which implements extractor for Hypertext Transfer
Protocol (HTTP/2), whose structure is described as below.

+-----------------------------------------------+
|                 Length (24)                   |
+---------------+---------------+---------------+
|   Type (8)    |   Flags (8)   |
+-+-------------+---------------+-------------------------------+
|R|                 Stream Identifier (31)                      |
+=+=============================================================+
|                   Frame Payload (0...)                      ...
+---------------------------------------------------------------+

"""
import collections

from jspcap.corekit.infoclass import Info
from jspcap.protocols.application.http import HTTP
from jspcap.utilities.exceptions import ProtocolError


__all__ = ['HTTPv2']


# HTTP/2 Types
_HTTP_TYPE = {
    0x00 : 'DATA',
    0x01 : 'HEADERS',
    0x02 : 'PRIORITY',
    0x03 : 'RST_STREAM',
    0x04 : 'SETTINGS',
    0x05 : 'PUSH_PROMISE',
    0x06 : 'PING',
    0x07 : 'GOAWAY',
    0x08 : 'WINDOW_UPDATE',
    0x09 : 'CONTINUATION',
}


# HTTP/2 Functions
_read_http_none = lambda self, size, kind, flag: self._read_http_none(size, kind, flag)     # Unsigned
_HTTP_FUNC = collections.defaultdict(lambda : _read_http_none, {
    0x00 : lambda self, size, kind, flag: self._read_http_data(size, kind, flag),           # DATA
    0x01 : lambda self, size, kind, flag: self._read_http_headers(size, kind, flag),        # HEADERS
    0x02 : lambda self, size, kind, flag: self._read_http_priority(size, kind, flag),       # PRIORITY
    0x03 : lambda self, size, kind, flag: self._read_http_rst_stream(size, kind, flag),     # RST_STREAM
    0x04 : lambda self, size, kind, flag: self._read_http_settings(size, kind, flag),       # SETTINGS
    0x05 : lambda self, size, kind, flag: self._read_http_push_promise(size, kind, flag),   # PUSH_PROMISE
    0x06 : lambda self, size, kind, flag: self._read_http_ping(size, kind, flag),           # PING
    0x07 : lambda self, size, kind, flag: self._read_http_goaway(size, kind, flag),         # GOAWAY
    0x08 : lambda self, size, kind, flag: self._read_http_window_update(size, kind, flag),  # WINDOW_UPDATE
    0x09 : lambda self, size, kind, flag: self._read_http_continuation(size, kind, flag),   # CONTINUATION
})


# Error Codes
_ERROR_CODE = {
    0x00 : 'NO_ERROR',
    0x01 : 'PROTOCOL_ERROR',
    0x02 : 'INTERNAL_ERROR',
    0x03 : 'FLOW_CONTROL_ERROR',
    0x04 : 'SETTINGS_TIMEOUT',
    0x05 : 'STREAM_CLOSED',
    0x06 : 'FRAME_SIZE_ERROR',
    0x07 : 'REFUSED_STREAM',
    0x08 : 'CANCEL',
    0x09 : 'COMPRESSION_ERROR',
    0x0a : 'CONNECT_ERROR',
    0x0b : 'ENHANCE_YOUR_CALM',
    0x0c : 'INADEQUATE_SECURITY',
    0x0d : 'HTTP_1_1_REQUIRED',
}


# SETTINGS Parameters
_PRAR_NAME = {
    0x01 : 'SETTINGS_HEADER_TABLE_SIZE',
    0x02 : 'SETTINGS_ENABLE_PUSH',
    0x03 : 'SETTINGS_MAX_CONCURRENT_STREAMS',
    0x04 : 'SETTINGS_INITIAL_WINDOW_SIZE',
    0x05 : 'SETTINGS_MAX_FRAME_SIZE',
    0x06 : 'SETTINGS_MAX_HEADER_LIST_SIZE',
}


class HTTPv2(HTTP):
    """This class implements Hypertext Transfer Protocol (HTTP/2).

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Application`
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_http -- read Hypertext Transfer Protocol (HTTP/2)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _make_protochain -- make ProtoChain instance for corresponding protocol
        * _read_http_data -- read HTTP/2 DATA frames
        * _read_http_headers -- read HTTP/2 HEADERS frames
        * _read_http_priority -- read HTTP/2 PRIORITY frames
        * _read_http_rst_stream -- read HTTP/2 RST_STREAM frames
        * _read_http_settings -- read HTTP/2 SETTINGS frames
        * _read_http_push_promise -- read HTTP/2 PUSH_PROMISE frames
        * _read_http_ping -- read HTTP/2 PING frames
        * _read_http_goaway -- read HTTP/2 GOAWAY frames
        * _read_http_window_update -- read HTTP/2 WINDOW_UPDATE frames
        * _read_http_continuation -- read HTTP/2 CONTINUATION frames

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def alias(self):
        """Acronym of current protocol."""
        return 'HTTP/2'

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_http(self, length):
        """Read Hypertext Transfer Protocol (HTTP/2).

        Structure of HTTP/2 packet [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +=+=============================================================+
            |                   Frame Payload (0...)                      ...
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.payload            Frame Payload

        """
        if length is None:
            length = len(self)

        if length < 9:
            raise ProtocolError(f'HTTP/2: invalid format', quiet=True)

        _tlen = self._read_unpack(3)
        _type = self._read_unpack(1)
        _flag = self._read_binary(1)
        _rsid = self._read_binary(4)

        if _tlen != length:
            raise ProtocolError(f'HTTP/2: [Type {_type}] invalid format', quiet=True)

        if int(_rsid[0], base=2):
            raise ProtocolError(f'HTTP/2: [Type {_type}] invalid format', quiet=True)

        http = dict(
            length = _tlen,
            type = _HTTP_TYPE.get(_type),
            sid = int(_rsid[1:], base=2),
            packet = self._read_packet(_tlen),
        )

        if http['type'] is None:
            raise ProtocolError(f'HTTP/2: [Type {_type}] invalid format', quiet=True)

        if http['type'] in ('SETTINGS', 'PING') and http['sid'] != 0:
            raise ProtocolError(f'HTTP/2: [Type {_type}] invalid format', quiet=True)

        _http = _HTTP_FUNC[_type](self, _tlen, _type, _flag)
        http.update(_http)

        return http

    ##########################################################################
    # Data models.
    ##########################################################################

    def __length_hint__(self):
        return 9

    @classmethod
    def __index__(cls):
        return cls.__name__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_http_none(self, size, kind, flag):
        """Read HTTP packet with unsigned type."""
        if any((int(bit, base=2) for bit in flag)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = None,
            payload = self._read_fileng(size - 9) or None,
        )

        return data

    def _read_http_data(self, size, kind, flag):
        """Read HTTP/2 DATA frames.

        Structure of HTTP/2 DATA frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------+-----------------------------------------------+
            |Pad Length? (8)|
            +---------------+-----------------------------------------------+
            |                            Data (*)                         ...
            +---------------------------------------------------------------+
            |                           Padding (*)                       ...
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (0)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.pad_len            Pad Length (Optional)
              10         80     http.data               Data
              ?           ?     -                       Padding (Optional)

        """
        _plen = 0
        _flag = dict(
            END_STREAM = False, # bit 0
            PADDED = False,     # bit 3
        )
        for index, bit in enumerate(flag):
            if index == 0 and bit:
                _flag['END_STREAM'] = True
            elif index == 3 and bit:
                _flag['PADDED'] = True
                _plen = self._read_unpack(1)
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        if _plen > size - 10:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        if _flag['PADDED']:
            _dlen = size - _plen - 1
        else:
            _dlen = size - _plen
        if _dlen < 0:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _data = self._read_fileng(_dlen)

        padding = self._read_binary(_plen)
        if any((int(bit, base=2) for bit in padding)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = _flag,
            data = _data,
        )
        if _flag['PADDED']:
            data['ped_len'] = _plen

        return data

    def _read_http_headers(self, size, kind, flag):
        """Read HTTP/2 HEADERS frames.

        Structure of HTTP/2 HEADERS frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------+-----------------------------------------------+
            |Pad Length? (8)|
            +-+-------------+-----------------------------------------------+
            |E|                 Stream Dependency? (31)                     |
            +-+-------------+-----------------------------------------------+
            |  Weight? (8)  |
            +-+-------------+-----------------------------------------------+
            |                   Header Block Fragment (*)                 ...
            +---------------------------------------------------------------+
            |                           Padding (*)                       ...
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (1)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.pad_len            Pad Length (Optional)
              10         80     http.exclusive          Exclusive Flag
              10         81     http.deps               Stream Dependency (Optional)
              14        112     http.weight             Weight (Optional)
              15        120     http.frag               Header Block Fragment
              ?           ?     -                       Padding (Optional)

        """
        _plen = 0
        _elen = 0
        _flag = dict(
            END_STREAM = False,     # bit 0
            END_HEADERS = False,    # bit 2
            PADDED = False,         # bit 3
            PRIORITY = False,       # bit 5
        )
        for index, bit in enumerate(flag):
            if index == 0 and bit:
                _flag['END_STREAM'] = True
            elif index == 2 and bit:
                _flag['END_HEADERS'] = True
            elif index == 3 and bit:
                _flag['PADDED'] = True
                _plen = self._read_unpack(1)
            elif index == 5 and bit:
                _flag['PRIORITY'] = True
                _edep = self._read_binary(4)
                _wght = self._read_unpack(1)
                _elen = 5
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        if _flag['PADDED']:
            _dlen = size - _plen - _elen - 1
        else:
            _dlen = size - _plen - _elen
        if _dlen < 0:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _frag = self._read_fileng(_dlen) or None

        padding = self._read_binary(_plen)
        if any((int(bit, base=2) for bit in padding)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = _flag,
            frag = _frag,
        )
        if _flag['PADDED']:
            data['ped_len'] = _plen
        if _flag['PRIORITY']:
            data['exclusive'] = True if int(_edep[0], base=2) else False
            data['deps'] = int(_edep[1:], base=2)
            data['weight'] = _wght + 1

        return data

    def _read_http_priority(self, size, kind, flag):
        """Read HTTP/2 PRIORITY frames.

        Structure of HTTP/2 PRIORITY frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +-+-------------------------------------------------------------+
            |E|                  Stream Dependency (31)                     |
            +-+-------------+-----------------------------------------------+
            |   Weight (8)  |
            +-+-------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.exclusive          Exclusive Flag
              9          73     http.deps               Stream Dependency
              13        104     http.weight             Weight

        """
        if size != 9:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
        if any((int(bit, base=2) for bit in flag)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _edep = self._read_binary(4)
        _wght = self._read_unpack(1)

        data = dict(
            flags = None,
            exclusive = True if int(_edep[0], base=2) else False,
            deps = int(_edep[1:], base=2),
            weight = _wght + 1,
        )

        return data

    def _read_http_rst_stream(self, size, kind, flag):
        """Read HTTP/2 RST_STREAM frames.

        Structure of HTTP/2 RST_STREAM frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------------------------------------------------------+
            |                        Error Code (32)                        |
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.error              Error Code

        """
        if size != 8:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
        if any((int(bit, base=2) for bit in flag)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _code = self._read_unpack(4)

        data = dict(
            flags = None,
            error = _ERROR_CODE.get(_code, _code),
        )

        return data

    def _read_http_settings(self, size, kind, flag):
        """Read HTTP/2 SETTINGS frames.

        Structure of HTTP/2 SETTINGS frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------------------------------------------------------+
            |       Identifier (16)         |
            +-------------------------------+-------------------------------+
            |                        Value (32)                             |
            +---------------------------------------------------------------+
            |                          ......                               |

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.settings           Settings
              9          72     http.settings.id        Identifier
              10         80     http.settings.value     Value

        """
        if size % 5 != 0:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _flag = dict(
            ACK = False,    # bit 0
        )
        for index, bit in enumerate(flag):
            if index == 0 and bit:
                _flag['ACK'] = True
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        if _flag['ACK'] and size:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _para = dict()
        counter = 0
        while counter < size:
            _stid = self._read_unpack(1)
            _pval = self._read_unpack(4)
            _pkey = _PRAR_NAME.get(_stid, 'Unsigned')
            if _pkey in _para:
                if isinstance(_para[_pkey], tuple):
                    _para[_pkey] += (_pval,)
                else:
                    _para[_pkey] = (_para[_pkey], _pval)
            else:
                _para[_pkey] = _pval

        data = dict(
            flags = _flag,
        )
        data.update(_para)

        return data

    def _read_http_push_promise(self, size, kind, flag):
        """Read HTTP/2 PUSH_PROMISE frames.

        Structure of HTTP/2 PUSH_PROMISE frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------+-----------------------------------------------+
            |Pad Length? (8)|
            +-+-------------+-----------------------------------------------+
            |R|                  Promised Stream ID (31)                    |
            +-+-----------------------------+-------------------------------+
            |                   Header Block Fragment (*)                 ...
            +---------------------------------------------------------------+
            |                           Padding (*)                       ...
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (1)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.pad_len            Pad Length (Optional)
              10         80     -                       Reserved
              10         81     http.pid                Promised Stream ID
              14        112     http.frag               Header Block Fragment
              ?           ?     -                       Padding (Optional)

        """
        if size < 4:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _plen = 0
        _flag = dict(
            END_HEADERS = False,    # bit 2
            PADDED = False,         # bit 3
        )
        for index, bit in enumerate(flag):
            if index == 2 and bit:
                _flag['END_HEADERS'] = True
            elif index == 3 and bit:
                _flag['PADDED'] = True
                _plen = self._read_unpack(1)
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        if _flag['PADDED']:
            _dlen = size - _plen - 5
        else:
            _dlen = size - _plen - 4
        if _dlen < 0:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _rpid = self._read_binary(4)
        _frag = self._read_fileng(_dlen) or None

        if int(_rpid[0], base=2):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        padding = self._read_binary(_plen)
        if any((int(bit, base=2) for bit in padding)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = _flag,
            pid = int(_rpid[1:], base=2),
            frag = _frag,
        )
        if _flag['PADDED']:
            data['ped_len'] = _plen

        return data

    def _read_http_ping(self, size, kind, flag):
        """Read HTTP/2 PING frames.

        Structure of HTTP/2 PING frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------------------------------------------------------+
            |                                                               |
            |                      Opaque Data (64)                         |
            |                                                               |
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     http.data               Opaque Data

        """
        if size != 8:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _flag = dict(
            ACK = False,    # bit 0
        )
        for index, bit in enumerate(flag):
            if index == 0 and bit:
                _flag['ACK'] = True
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        _data = self._read_fileng(8)

        data = dict(
            flags = _flag,
            data = _data,
        )

        return data

    def _read_http_goaway(self, size, kind, flag):
        """Read HTTP/2 GOAWAY frames.

        Structure of HTTP/2 GOAWAY frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +-+-------------+---------------+-------------------------------+
            |R|                  Last-Stream-ID (31)                        |
            +-+-------------------------------------------------------------+
            |                      Error Code (32)                          |
            +---------------------------------------------------------------+
            |                  Additional Debug Data (*)                    |
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     -                       Reserved
              9          73     http.last_sid           Last Stream ID
              13        104     http.error              Error Code
              17        136     http.data               Additional Debug Data (Optional)

        """
        _dlen = size - 8
        if _dlen < 0:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
        if any((int(bit, base=2) for bit in flag)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _rsid = self._read_binary(4)
        _code = self._read_unpack(4)
        _data = self._read_fileng(_dlen) or None

        if int(_rsid[0], base=2):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = None,
            last_sid = int(_rsid[1:], base=2),
            error = _ERROR_CODE.get(_code, _code),
            data = _data,
        )

        return data

    def _read_http_window_update(self, size, kind, flag):
        """Read HTTP/2 WINDOW_UPDATE frames.

        Structure of HTTP/2 WINDOW_UPDATE frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +-+-------------+---------------+-------------------------------+
            |R|              Window Size Increment (31)                     |
            +-+-------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          72     -                       Reserved
              9          73     http.window             Window Size Increment

        """
        if size != 4:
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
        if any((int(bit, base=2) for bit in flag)):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        _size = self._read_binary(4)

        if int(_size[0], base=2):
            raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)

        data = dict(
            flags = None,
            window = int(_size[1:], base=2),
        )

        return data

    def _read_http_continuation(self, size, kind, flag):
        """Read HTTP/2 WINDOW_UPDATE frames.

        Structure of HTTP/2 WINDOW_UPDATE frame [RFC 7540]:
            +-----------------------------------------------+
            |                 Length (24)                   |
            +---------------+---------------+---------------+
            |   Type (8)    |   Flags (8)   |
            +-+-------------+---------------+-------------------------------+
            |R|                 Stream Identifier (31)                      |
            +---------------------------------------------------------------+
            |                   Header Block Fragment (*)                 ...
            +---------------------------------------------------------------+

            Octets      Bits        Name                    Discription
              0           0     http.length             Length
              3          24     http.type               Type (2)
              4          32     http.flags              Flags
              5          40     -                       Reserved
              5          41     http.sid                Stream Identifier
              9          73     http.frag               Header Block Fragment

        """
        _flag = dict(
            END_HEADERS = False,    # bit 2
        )
        for index, bit in enumerate(flag):
            if index == 2 and bit:
                _flag['END_HEADERS'] = True
            elif bit:
                raise ProtocolError(f'HTTP/2: [Type {kind}] invalid format', quiet=True)
            else:
                continue

        _frag = self._read_fileng(size) or None

        data = dict(
            flags = _flag,
            frag = _frag,
        )

        return data



# -*- coding: utf-8 -*-
"""

"""
import time

from jspcap.ipsuite.protocol import Protocol


__all__ = ['Frame']


class Frame(Protocol):
    """PCAP frame header constructor.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * data -- bytes, binary packet data if current instance
        * alias -- str, acronym of corresponding protocol

    Methods:
        * index -- return first index of value from a dict
        * pack -- pack integers to bytes
        * update -- update packet data

    """
    ##########################################################################
    # Methods.
    ##########################################################################

    def update(self, **kwargs):
        """Update packet data."""
        # update dict
        self.__dict__.update(kwargs)

        # fetch values
        packet = self.__dict__.pop('packet', bytes())               # raw packet data
        frame_info = self.__dict__.pop('frame_info')                # extracted frame info
        if frame_info and isinstance(frame_info, dict):
            now = [ int(x or 0) for x in str(time.time()).split('.') ]
            ts_sec = frame_info.pop('ts_sec', now[0])               # timestamp seconds
            ts_usec = frame_info.pop('ts_usec', now[1])             # timestamp microseconds
            incl_len = frame_info.pop('incl_len', len(packet))      # number of octets of packet saved in file
            orig_len = frame_info.pop('orig_len', len(packet))      # actual length of packet
        else:
            now = [ int(x or 0) for x in str(time.time()).split('.') ]
            ts_sec = self.__dict__.pop('ts_sec', now[0])            # timestamp seconds
            ts_usec = self.__dict__.pop('ts_usec', now[1])          # timestamp microseconds
            incl_len = self.__dict__.pop('incl_len', len(packet))   # number of octets of packet saved in file
            orig_len = self.__dict__.pop('orig_len', len(packet))   # actual length of packet

        # update packet
        data = self.pack(ts_sec, size=4, lilendian=True)
        data += self.pack(ts_usec, size=4, lilendian=True)
        data += self.pack(incl_len, size=4, lilendian=True)
        data += self.pack(orig_len, size=4, lilendian=True)
        data += packet

        # update data
        self.__data__ = data



# -*- coding: utf-8 -*-
"""version info

`jspcap.corekit.version` contains tuple-like class
`VersionInfo`, which is originally designed alike
`sys.version_info`.

"""
import collections


__all__ = ['VersionInfo']


VersionInfo = collections.namedtuple('VersionInfo', ['major', 'minor'])
VersionInfo.__doc__ = 'VersionInfo is alike `sys.version_info`.'



# -*- coding: utf-8 -*-
"""root protocol

`jspcap.ipsuite.protocol` contains `Protocol` only,
which is an abstract base clss for all protocol family,
with pre-defined utility arguments and methods of specified
protocols.

"""
import abc
import struct

from jspcap.corekit.infoclass import Info
from jspcap.utilities.exceptions import ProtocolNotImplemented, StructError
from jspcap.utilities.validations import dict_check


__all__ = ['Protocol']


# abstract base class utilities
ABCMeta = abc.ABCMeta
abstractmethod = abc.abstractmethod
abstractproperty = abc.abstractproperty


class Protocol:
    """Abstract base class for Internet Protocol Suite.

    Properties:
        * name -- str, name of corresponding protocol
        * info -- Info, info dict of current instance
        * data -- bytes, binary packet data if current instance
        * alias -- str, acronym of corresponding protocol

    Methods:
        * index -- return first index of value from a dict
        * pack -- pack integers to bytes
        * update -- update packet data

    """
    __metaclass__ = ABCMeta

    ##########################################################################
    # Properties.
    ##########################################################################

    # name of current protocol
    @abstractproperty
    def name(self):
        """Name of current protocol."""
        pass

    # acronym of current protocol
    @property
    def alias(self):
        """Acronym of current protocol."""
        return self.__class__.__name__

    # info dict of current instance
    @property
    def info(self):
        """Info dict of current instance."""
        return Info(self.__dict__)

    # binary packet data if current instance
    @property
    def data(self):
        """Binary packet data if current instance."""
        return self.__data__

    ##########################################################################
    # Methods.
    ##########################################################################

    @classmethod
    def index(cls, dict_, value, *, pack=False, size=4, lilendian=False):
        """Return first index of value from a dict."""
        try:
            index = list(dict_.keys())[list(dict_.values()).index(value)]
            if pack:
                return cls.pack(index, size=size, lilendian=lilendian)
            return index
        except ValueError:
            raise ProtocolNotImplemented(f'protocol {value} not implemented') from None

    @abstractmethod
    def update(self, **kwargs):
        """Update packet data."""
        pass

    @staticmethod
    def pack(integer, *, size=1, signed=False, lilendian=False):
        """Pack integers to bytes.

        Positional arguments:
            * integer  -- int, integer to be packed

        Keyword arguments:
            * size  -- int, buffer size (default is 1)
            * signed -- bool, signed flag (default is False)
                           <keyword> True / False
            * lilendian -- bool, little-endian flag (default is False)
                           <keyword> True / False

        Returns:
            * bytes -- packed data upon success

        """
        endian = '<' if lilendian else '>'
        if size == 8:   kind = 'q' if signed else 'Q'   # unpack to 8-byte integer (long long)
        elif size == 4: kind = 'i' if signed else 'I'   # unpack to 4-byte integer (int / long)
        elif size == 2: kind = 'h' if signed else 'H'   # unpack to 2-byte integer (short)
        elif size == 1: kind = 'b' if signed else 'B'   # unpack to 1-byte integer (char)
        else:           kind = None                     # do not unpack

        if kind is None:
            end = 'little' if lilendian else 'big'
            buf = integer.to_bytes(size, end, signed=signed)
        else:
            try:
                fmt = f'{endian}{kind}'
                buf = struct.pack(fmt, integer)
            except struct.error:
                raise StructError(f'{self.__class__.__name__}: pack failed') from None
        return buf

    ##########################################################################
    # Data models.
    ##########################################################################

    # Not hashable
    __hash__ = None

    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls)
        return self

    def __init__(self, *args, **kwargs):
        self.__dict__ = kwargs
        for arg in args:
            if isinstance(arg, Info):
                arg = arg.infotodict()
            dict_check(arg)
            self.__dict__.update(arg)
        self.update()

    @classmethod
    def __index__(cls):
        return cls.__name__



# -*- coding: utf-8 -*-
"""destination options for IPv6

`jspcap.protocols.internet.ipv6_opts` contains
`IPv6_Opts` only, which implements extractor for
Destination Options for IPv6 (IPv6-Opts), whose structure
is described as below.

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|  Next Header  |  Hdr Ext Len  |                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               +
|                                                               |
.                                                               .
.                            Options                            .
.                                                               .
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

"""
# TODO: Implements extractor of all destination options.
from jspcap.corekit.infoclass import Info
from jspcap.protocols.protocol import Protocol


__all__ = ['IPv6_Opts']


class IPv6_Opts(Protocol):
    """This class implements Destination Options for IPv6.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Methods:
        * read_ipv6_opts -- read Destination Options for IPv6 (IPv6-Opts)

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    ##########################################################################
    # Properties.
    ##########################################################################

    @property
    def name(self):
        """Name of current protocol."""
        return 'IPv6 Hop-by-Hop Options'

    @property
    def alias(self):
        """Acronym of corresponding procotol."""
        return 'IPv6-Opts'

    @property
    def length(self):
        """Header length of current protocol."""
        return self._info.length

    @property
    def protocol(self):
        """Name of next layer protocol."""
        return self._info.next

    ##########################################################################
    # Methods.
    ##########################################################################

    def read_ipv6_opts(self, length, extension):
        """Read Destination Options for IPv6.

        Structure of IPv6-Opts header [RFC 8200]:
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
            |  Next Header  |  Hdr Ext Len  |                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+                               +
            |                                                               |
            .                                                               .
            .                            Options                            .
            .                                                               .
            |                                                               |
            +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

            Octets      Bits        Name                    Discription
              0           0     opt.next                Next Header
              1           8     opt.length              Header Extensive Length
              2          16     opt.options             Options

        """
        if length is None:
            length = len(self)

        _next = self._read_protos(1)
        _hlen = self._read_unpack(1)
        _opts = self._read_fileng(_hlen - 1)

        ipv6_opts = dict(
            next = _next,
            length = _hlen + 1,
            options = _opts,
        )

        if length is not None:
            length -= ipv6_opts['length']
        ipv6_opts['packet'] = self._read_packet(header=ipv6_opts['length'], payload=length)

        if extension:
            self._protos = None
            return ipv6_opts
        return self._decode_next_layer(ipv6_opts, _next, length)

    ##########################################################################
    # Data models.
    ##########################################################################

    def __init__(self, _file, length=None, *, extension=False, **kwargs):
        self._file = _file
        self._info = Info(self.read_ipv6_opts(length, extension))

    def __length_hint__(self):
        return 2



# -*- coding: utf-8 -*-
"""internet layer protocols

`jspcap.protocols.internet` is collection of all protocols
in internet layer, with detailed implementation and
methods.

"""
# TODO: Implements ECN, ESP, ICMP, ICMPv6, IGMP, Shim6.

# Base Class for Internet Layer
from jspcap.protocols.internet.internet import Internet

# Utility Classes for Protocols
from jspcap.protocols.internet.ah import AH
from jspcap.protocols.internet.ipv4 import IPv4
from jspcap.protocols.internet.ipv6 import IPv6
from jspcap.protocols.internet.ipx import IPX

# IPv6 Extension Headers
from jspcap.protocols.internet.hip import HIP
from jspcap.protocols.internet.hopopt import HOPOPT
from jspcap.protocols.internet.ipv6_frag import IPv6_Frag
from jspcap.protocols.internet.ipv6_opts import IPv6_Opts
from jspcap.protocols.internet.ipv6_route import IPv6_Route
from jspcap.protocols.internet.mh import MH

# Ethertype IEEE 802 Numbers
from jspcap.protocols.internet.internet import ETHERTYPE

# Deprecated / Base Classes
from jspcap.protocols.internet.ip import IP
from jspcap.protocols.internet.ipsec import IPsec


__all__ = [
    'ETHERTYPE',                                        # Protocol Numbers
    'AH', 'IP', 'IPsec', 'IPv4', 'IPv6', 'IPX',         # Internet Layer
    'HIP', 'HOPOPT', 'IPv6_Frag',
    'IPv6_Opts', 'IPv6_Route', 'MH',                    # IPv6 Extension Header
]



# -*- coding: utf-8 -*-
"""root internet layer protocol

`jspcap.protocols.internet.internet` contains both
`ETHERTYPE` and `Internet`. The former is a dictionary
of ethertype IEEE 802 numbers, registered in IANA. And the
latter is a base class for internet layer protocols, eg.
AH, IP, IPsec, IPv4, IPv6, IPX, and etc.

"""
import io

from jspcap.corekit.protochain import ProtoChain
from jspcap.protocols.protocol import Protocol
from jspcap.protocols.transport.transport import TP_PROTO
from jspcap.utilities.decorators import beholder


__all__ = ['Internet', 'ETHERTYPE']


# Ethertype IEEE 802 Numbers
ETHERTYPE = {
    # Link Layer
    0x0806 : 'ARP',     # Address Resolution Protocol
    0x8035 : 'RARP',    # Reverse Address Resolution Protocol
    0x8100 : 'VLAN',    # 802.1Q Customer VLAN Tag Type

    # Internet Layer
    0x0800 : 'IPv4',    # Internet Protocol version 4
    0x8137 : 'IPX',     # Internetwork Packet Exchange
    0x86dd : 'IPv6',    # Internet Protocol version 6
}


class Internet(Protocol):
    """Abstract base class for internet layer protocol family.

    Properties:
        * name -- str, name of corresponding procotol
        * info -- Info, info dict of current instance
        * alias -- str, acronym of corresponding procotol
        * layer -- str, `Internet`
        * length -- int, header length of corresponding protocol
        * protocol -- str, name of next layer protocol
        * protochain -- ProtoChain, protocol chain of current instance

    Attributes:
        * _file -- BytesIO, bytes to be extracted
        * _info -- Info, info dict of current instance
        * _protos -- ProtoChain, protocol chain of current instance

    Utilities:
        * _read_protos -- read next layer protocol type
        * _read_fileng -- read file buffer
        * _read_unpack -- read bytes and unpack to integers
        * _read_binary -- read bytes and convert into binaries
        * _read_packet -- read raw packet data
        * _decode_next_layer -- decode next layer protocol type
        * _import_next_layer -- import next layer protocol extractor

    """
    __layer__ = 'Internet'

    ##########################################################################
    # Properties.
    ##########################################################################

    # protocol layer
    @property
    def layer(self):
        """Protocol layer."""
        return self.__layer__

    ##########################################################################
    # Utilities.
    ##########################################################################

    def _read_protos(self, size):
        """Read next layer protocol type.

        Positional arguments:
            * size  -- int, buffer size

        Returns:
            * str -- next layer's protocol name

        """
        _byte = self._read_unpack(size)
        _prot = TP_PROTO.get(_byte)
        return _prot

    def _decode_next_layer(self, dict_, proto=None, length=None, *, version=4):
        """Decode next layer extractor.

        Positional arguments:
            * dict_ -- dict, info buffer
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Keyword Arguments:
            * version -- int, IP version (4 in default)
                            <keyword> 4 / 6

        Returns:
            * dict -- current protocol with next layer extracted

        """
        if self._onerror:
            flag, info, chain, alias = beholder(self._import_next_layer)(self, proto, length, version=version)
        else:
            flag, info, chain, alias = self._import_next_layer(proto, length, version=version)

        # make next layer protocol name
        if flag:
            if proto is None and chain:
                layer = chain.alias[0].lower()
                proto, chain = chain.tuple[0], None
            else:
                layer = str(alias or proto or 'Raw').lower()
        else:
            layer, proto = 'raw', 'Raw'

        # write info and protocol chain into dict
        dict_[layer] = info
        self._protos = ProtoChain(proto, chain, alias)
        return dict_

    def _import_next_layer(self, proto, length=None, *, version=4, extension=False):
        """Import next layer extractor.

        Positional arguments:
            * proto -- str, next layer protocol name
            * length -- int, valid (not padding) length

        Keyword Arguments:
            * version -- int, IP version (4 in default)
                            <keyword> 4 / 6
            * extension -- bool, if is extension header (False in default)
                            <keyword> True / False

        Returns:
            * bool -- flag if extraction of next layer succeeded
            * Info -- info of next layer
            * ProtoChain -- protocol chain of next layer
            * str -- alias of next layer

        Protocols:
            * IPv4 -- internet layer
            * IPv6 -- internet layer
            * AH -- internet layer
            * TCP -- transport layer
            * UDP -- transport layer

        """
        if self._sigterm:
            from jspcap.protocols.raw import Raw as Protocol
        elif proto == 'AH':
            from jspcap.protocols.internet.ah import AH as Protocol
        elif proto == 'HIP':
            from jspcap.protocols.internet.hip import HIP as Protocol
        elif proto == 'HOPOPT':
            from jspcap.protocols.internet.hopopt import HOPOPT as Protocol
        elif proto == 'IPv6-Frag':
            from jspcap.protocols.internet.ipv6_frag import IPv6_Frag as Protocol
        elif proto == 'IPv6-Opts':
            from jspcap.protocols.internet.ipv6_opts import IPv6_Opts as Protocol
        elif proto == 'IPv6-Route':
            from jspcap.protocols.internet.ipv6_route import IPv6_Route as Protocol
        elif proto == 'MH':
            from jspcap.protocols.internet.mh import MH as Protocol
        elif proto == 'IPv4':
            from jspcap.protocols.internet.ipv4 import IPv4 as Protocol
        elif proto == 'IPv6':
            from jspcap.protocols.internet.ipv6 import IPv6 as Protocol
        elif proto == 'TCP':
            from jspcap.protocols.transport.tcp import TCP as Protocol
        elif proto == 'UDP':
            from jspcap.protocols.transport.udp import UDP as Protocol
        else:
            from jspcap.protocols.raw import Raw as Protocol
        next_ = Protocol(io.BytesIO(self._read_fileng(length)), length,
                            version=version, extension=extension,
                            error=self._onerror, layer=self._exlayer, protocol=self._exproto)
        return True, next_.info, next_.protochain, next_.alias
